// Copyright (C) 2015, Antoni Buades, Universitat Illes Balears
// // <toni.buades@uib.es>
// //
// //
// //
// // This program is free software: you can use, modify and/or
// // redistribute it under the terms of the GNU General Public
// // License as published by the Free Software Foundation, either
// // version 3 of the License, or (at your option) any later
// // version. You should have received a copy of this license along
// // this program. If not, see <http://www.gnu.org/licenses/>.
//
//
//


#include "libBasic.h"


namespace libUSTG
{



void fpCombine(float *u,float a,float *v,float b, float *w,  int size)
{
    for(int i=0; i<size ; i++)   w[i]= (a *  u[i] + b* v[i]);

}
////
////! Image convolutions
////

///**
// *
// * Neumann boundary condition test
// *
// **/
int neumann_bc(int x, int nx, bool *out)
{
    if(x < 0) {
        x = 0;
        *out = true;
    } else if (x >= nx) {
        x = nx - 1;
        *out = true;
    }

    return x;
}


/**
 *
 * Symmetric boundary condition test
 *
 **/
int symmetric_bc(int x, int nx, bool *out)
{
    if(x < 0) {
        const int borde = nx - 1;
        const int xx = -x;
        const int n  = (int)(xx/borde) % 2;

        if ( n ) x = borde - ( xx % borde );
        else x = xx % borde;
        *out = true;
    } else if ( x >= nx ) {
        const int borde = nx - 1;
        const int n = (int)(x/borde) % 2;

        if ( n ) x = borde - ( x % borde );
        else x = x % borde;
        *out = true;
    }

    return x;
}





float*  fiFloatGaussKernel(float std, int & size)
{



    int n = 4 * ceilf(std) + 1;
    size = n;


    float* u = new float[n];


    if (n==1)  u[0]=1.0;
    else {

        int ishift = (n-1) / 2;

        for (int i=ishift; i < n; i++) {

            float v = (float)(i - ishift) / std;

            u[i] = u[n-1-i] = (float) exp(-0.5*v*v);

        }

    }


    // normalize
    float fSum = 0.0f;
    for (int i=0; i < n; i++) fSum += u[i];
    for (int i=0; i < n; i++)  u[i] /= fSum;


    return u;

}



float * fiFloatDirectionalGaussKernel(float xsigma, float ysigma, float angle,
                                      float *kernel, int kwidth, int kheight)
{

    assert(kwidth == kheight);
    int ksize = kwidth;


    float xsigma2 = xsigma*xsigma;
    float ysigma2 = ysigma*ysigma;

    int l2 = ksize/2;
    for(int y = -l2; y <= l2; y++)
        for(int x = -l2; x <= l2; x++) {

            float a = (float) angle * PI / 180.0f;
            float sina = sin(a);
            float cosa = cos(a);

            float ax = (float) x * cosa + (float) y * sina;
            float ay = -(float) x * sina + (float) y * cosa;
            kernel[(y+l2) * ksize + x + l2] =  exp(-(ax*ax)/(2.0f*xsigma2) - \
                                                   (ay*ay)/(2.0f*ysigma2) );

        }


    float sum=0.0;
    for(int i=0; i < ksize*ksize; i++) sum += kernel[i];
    for(int i=0; i < ksize*ksize; i++) kernel[i] /= sum;

    kwidth = ksize;
    kheight = ksize;

    return kernel;
}


float * fiFloatDirectionalGaussKernelS(float xsigma, float ysigma, float angle,
                                       float *kernel, int kwidth, int kheight,
                                       int sign)
{

    assert(kwidth == kheight);
    int ksize = kwidth;


    float xsigma2 = xsigma*xsigma;
    float ysigma2 = ysigma*ysigma;

    int l2 = ksize/2;
    for(int y = -l2; y <= l2; y++)
        for(int x = -l2; x <= l2; x++) {

            float a = (float) angle * PI / 180.0f;
            float sina = sin(a);
            float cosa = cos(a);

            float ax = (float) x * cosa + (float) y * sina;
            float ay = -(float) x * sina + (float) y * cosa;

            float d2=x*x+y*y;

            if (ax*sign > 0)
                kernel[(y+l2) * ksize + x + l2] =  exp(-d2/(2.0f*ysigma2)  - \
                                                       (ax*ax)/(2.0f*xsigma2) );
            else
                kernel[(y+l2) * ksize + x + l2] =  exp(-d2/(2.0f*ysigma2) );


        }


    float sum=0.0;
    for(int i=0; i < ksize*ksize; i++) sum += kernel[i];
    for(int i=0; i < ksize*ksize; i++) kernel[i] /= sum;

    kwidth = ksize;
    kheight = ksize;

    return kernel;
}

void fiConvol_skernel(float *u,float *v,int width,int height,
                      struct sorted_kernel *skernel, int boundary)
{

    int K2 = skernel->w / 2;
    int L2 = skernel->h / 2;
    bool b;
    int l, k;
    int px, py;

    for(int y=0 ; y < height; y++)
        for (int x=0 ; x < width; x++) {

            float S = 0.0;

            if (boundary == BOUNDARY_CONDITION_NEUMANN) {
                for (int n=0; n < skernel->ncut; n++) {
                    l=skernel->items[n].j-L2;
                    k=skernel->items[n].i-K2;
                    px= neumann_bc(x+k, width, &b);
                    py= neumann_bc(y+l, height, &b);
                    S += skernel->items[n].v * u[width*py + px];
                }

            } else {

                for (int n=0; n < skernel->ncut; n++) {
                    l=skernel->items[n].j-L2;
                    k=skernel->items[n].i-K2;
                    px= symmetric_bc(x+k, width, &b);
                    py= symmetric_bc(y+l, height, &b);
                    S += skernel->items[n].v * u[width*py + px];
                }

            }

            v[y*width+x] = (float) S;

        }
}


////
//! Gradient based
//

void fiComputeImageGradient(float * fpI,float *fpXgrad, float *fpYgrad,
                            float *fpGrad, float * fpOri,
                            int iWidth, int iHeight, char cType)
{

    assert(fpI != NULL);
    assert(cType == 'c' || cType == 'f');

    int iC, iN, iS, iW, iE;
    float xgrad, ygrad;

    for (int ih = 0; ih < iHeight; ih++)
        for (int iw = 0; iw < iWidth; iw++) {

            //! Indexos
            iC = ih * iWidth + iw;
            iN = iC - iWidth;
            iS = iC + iWidth;
            iW = iC - 1;
            iE = iC + 1;

            //! Boundary
            if (ih == 0) iN = iC;
            if (ih == iHeight-1) iS = iC;
            if (iw == 0) iW = iC;
            if (iw == iWidth-1) iE = iC;



            if (cType == 'f') {
                //! Forward
                xgrad = fpI[iE] - fpI[iC];

                ygrad = fpI[iS] - fpI[iC] ;

            } else {

                //! Centered
                xgrad = fpI[iE] - fpI[iW];

                ygrad = fpI[iS] - fpI[iN];

            }


            if (fpXgrad) fpXgrad[ih * iWidth + iw] =  xgrad;

            if (fpYgrad) fpYgrad[ih * iWidth + iw] =  ygrad;

            if (fpGrad) fpGrad[ih * iWidth + iw] =  sqrtf(xgrad * xgrad + \
                                                          ygrad * ygrad);

            if (fpOri) fpOri[ih * iWidth + iw] =  atan2f(-ygrad,xgrad);

        }


}




void fiComputeImageGradient(float * fpI, float *fpGrad, float * fpOri,
                            int iWidth, int iHeight, char cType)
{
    fiComputeImageGradient( fpI, NULL, NULL, fpGrad,  fpOri, iWidth, iHeight,
                           cType);
}

void fiComputeImageGradient(float * fpI, float *fpGrad, int iWidth, int iHeight,
                            char cType)
{
    fiComputeImageGradient( fpI, NULL, NULL, fpGrad, NULL , iWidth, iHeight,
                           cType);

}



int order_decreasing_kernel(const void *pVoid1, const void *pVoid2)
{
    struct kernel_item *node1, *node2;

    node1=(struct kernel_item *) pVoid1;
    node2=(struct kernel_item *) pVoid2;

    if (node1->v < node2->v) return 1;
    if (node1->v > node2->v) return -1;
    return 0;
}

void sort_kernel(float *u, int w, int h, struct sorted_kernel *skernel,
                 float pkernel)
{
    for (int i=0; i < w; i++)
        for (int j=0; j < h; j++) {
            skernel->items[i+j*w].v=u[i+j*w];
            skernel->items[i+j*w].i=i;
            skernel->items[i+j*w].j=j;
        }

    qsort(skernel->items, skernel->nitems, sizeof(struct kernel_item), \
          order_decreasing_kernel);

    double S=0;
    int ncut=0;
    while ((S < pkernel) && (ncut < skernel->nitems)) {
        S+=(double) skernel->items[ncut++].v;
    }
    skernel->ncut=ncut;
    //renormalize values
    S=0;
    for (int n=0; n < ncut; n++) S+=(double) skernel->items[n].v;
    for (int n=0; n < ncut; n++) skernel->items[n].v/=S;

}






}




void  printusage(char *pname,
                 char *gp,
                 std::vector<OptStruct*>  &opt,
                 std::vector<ParStruct*>  &par)
{


    int npar = par.size();

    //! print function prototype
    //! "usage:" function_name [-r r] [-g g] [-b b]  par1 par2 ... parn
    printf("\nusage: %s ", pname);
    for(int i=0; i < (int) strlen(gp); i++)
        if (gp[i] != ':') {
            printf("[-%c",gp[i]);

            if (i+1 < (int) strlen(gp) && gp[i+1] ==  ':')
                printf(" %c] ", gp[i]);
            else printf("] ");

        }

    for(int i=0; i < npar; i++)
        printf(" %s ", par[i]->name);

    printf("\n");


    //! print options with associated descriptions and defaulted values
    int j=0;
    for(int i=0; i < (int) strlen(gp); i++)
        if (gp[i] != ':') {
            printf("\t-%c",gp[i]);

            if (i+1 < (int) strlen(gp) && gp[i+1] ==  ':') {

                printf("  %c\t %s ", gp[i], opt[j]->comment);
                if (opt[j]->defvalue != NULL)
                    printf("(Default: %s)",opt[j]->defvalue);

                printf("\n");

            } else printf("\t %s \n", opt[j]->comment);

            j++;
        }

    //! print mandatory parameters with associated descriptions
    for(int i=0; i < npar; i++) {
        printf("\t%s",par[i]->name);
        printf("\t %s\n", par[i]->comment);
    }


}




int parsecmdline(char *pname,
                 char *function,
                 int argc, char **argv,
                 std::vector <OptStruct*> & opt,
                 std::vector <ParStruct*> & par)
{

    //! number of options and obligatory parameters
    int nopt = opt.size();
    int npar = par.size();


    //! size of auxiliar parameter needed by getopt
    //! the length is at maximum 2*nopt since each option contains one caracter
    //! or two if the option requires a values
    char *gp = new char[2*nopt+1];
    gp[0]='\0';


    //! clear options and concatenate option identifiers into gp vector
    for(int i=0; i < nopt; i++) {
        opt[i]->flag = 0;
        opt[i]->value=NULL;
        strcat(gp, opt[i]->gp);
    }

    //! clear necessary parameter values
    for(int i=0; i < npar; i++) {
        par[i]->value = NULL;
    }


    //! in this way getopt doesn't print any information and errors are
    //! treated and printed by this program.
    opterr = 0;


    //! getopt return the following option of the concatenated vector gp
    //! or -1 if all options have already been treated
    //! it also fills "optarg" with the associated value to this option as
    //it is given in argv
    int c;
    while ((c = getopt (argc, argv, gp)) != -1) {

        //! set current option given by getopt
        //! if getopt finds an option which was not in the option of
        //the program (vector gp)
        //! returns ?
        int j=0;
        for(unsigned int i=0; i < strlen(gp); i++)
            if (c == gp[i]) {
                opt[j]->flag = 1;
                opt[j]->value = optarg;  // getopt fills "optarg" with the
              //associated value to this option as it is given in argv

                break;

            } else if (gp[i] != ':') j++;


        //! option found in console command was not one of the parameters of
        //our program
        //! or should have a mandatory values which is not provided
        if (c == '?') {


            //! when getopt encounters an unknown option character or an option
            //with a missing required argument
            //! it stores that option character in optopt variable
            unsigned int i = 0;
            for(i=0; i < strlen(gp); i++)
                if (optopt == gp[i]) {
                  printf("\n%s: %s\n", pname, function);
                  printf("\nerror: option -%c requires an argument.\n", optopt);
                  break;
                }

            if (i == strlen(gp)) {
                printf("\n%s: %s\n", pname, function);
                printf ("\nerror: unknown option `-%c'.\n", optopt);
            }

            printusage(pname, gp,  opt,  par);
            delete[] gp;
            return 0;

        }



    }


    //! Setting default values for non selected options
    for(int j=0; j < nopt; j++)
        if (opt[j]->flag == 0 && opt[j]->defvalue != NULL)
           opt[j]->value =  opt[j]->defvalue;


    //! Check remaining words in command after reading option
    if (argc - optind != npar) {
        printf("\n%s: %s\n", pname, function);
        fprintf (stderr, "\nerror: incorrect number of parameters\n");
        printusage(pname, gp,  opt,par);
        delete[] gp;
        return 0;
    }

    //! Read mandatory parameter values
    int i=0;
    for (int index = optind; index < argc ; index++, i++) {
        par[i]->value = argv[index];
    }


    delete[] gp;
    return 1;


}




