// Copyright (C) 2015, Antoni Buades, Universitat Illes Balears
// // <toni.buades@uib.es>
// //
// //
// //
// // This program is free software: you can use, modify and/or
// // redistribute it under the terms of the GNU General Public
// // License as published by the Free Software Foundation, either
// // version 3 of the License, or (at your option) any later
// // version. You should have received a copy of this license along
// // this program. If not, see <http://www.gnu.org/licenses/>.
//
//
//


#ifndef __library__
#define __library__


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <cmath>
#include <cmath>
#include <cassert>
#include <vector>
#include <getopt.h>


#include <fftw3.h>


namespace libUSTG
{


#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

#ifndef M_PI
#define M_PI 3.14159265358979323846264338327
#endif

#define dTiny 1e-10
#define fTiny 0.00000001f
#define fLarge 100000000.0f
#define dLarge 1e+10

#define MAX(i,j) ( (i)<(j) ? (j):(i) )
#define MIN(i,j) ( (i)<(j) ? (i):(j) )


struct kernel_item {
    float v;
    int i, j;
};

struct sorted_kernel {
    int nitems;
    int w, h;
    int ncut;
    struct kernel_item *items;
};



//! Forward declarations
class laMatrix;
class laVector;




/////////////////////////////////////////////////////////////////////
////
////! Float Value operations
////
/////////////////////////////////////////////////////////////////////////////
  
void fpCombine(float *u,float a,float *v,float b, float *w,  int size);

////////////////////////////////////////////////////////////////////////////////
////
////! Image Convolution
////
///////////////////////////////////////////////////////////////////////////
//
#define BOUNDARY_CONDITION_NEUMANN 0
#define BOUNDARY_CONDITION_SYMMETRIC 1
//
float* fiFloatGaussKernel(float std, int & size);
float* fiFloatDirectionalGaussKernel(float xsigma, float ysigma, float angle,
                                     float *kernel, int kwidth, int kheight);
float * fiFloatDirectionalGaussKernelS(float xsigma, float ysigma, float angle,
                                       float *kernel, int kwidth, int kheight,
                                       int sign);
void fiConvol_skernel(float *u,float *v,int width,int height,
                      struct sorted_kernel *skernel, int boundary);
  
void sort_kernel(float *u, int w, int h, struct sorted_kernel *skernel,
                 float pkernel);

//////////////////////////////////////////////////////////////
//  ////////////////////////////
////
////! Gradient Computation
////
///////////////////////////////////////////////////////////////
//  ////////////////////////////

void fiComputeImageGradient(float * fpI,float *fpXgrad, float *fpYgrad,
                            float *fpGrad, float * fpOri, int iWidth,
                            int iHeight, char cType = 'f');

void fiComputeImageGradient(float * fpI, float *fpGrad, float * fpOri,
                            int iWidth, int iHeight, char cType = 'f');

void fiComputeImageGradient(float * fpI, float *fpGrad, int iWidth,
                            int iHeight, char cType = 'f');

}

//
//! Parser
//

//! structure for parameters and options which are
//! optional in the code or they already have a default value
typedef struct optstruct {
    char *gp;           //! string of two letters  "a:"  as necessary
                        // for using getopt afterwards
    //! the ":" indicates that the activation of the option requires a value
    //          for the parameter
    //! and "a" that this option is activated by "-a " in the command

    int flag;           //! flag indicating that the option has been activated

    char *defvalue;     //! default value for this parameter if not modified
                        //  by console

    char *value;        //! value of the associated parameter to current option

    char *comment;      //! comment that appears by console

} OptStruct;



//! structure for necessary parameters of the method
typedef struct parstruct {
    char * name;
    char * value;       //! value of the parameter
    char * comment;     //! comment that appears by console

} ParStruct;



int parsecmdline(char *pname,
                 char *function,
                 int argc, char **argv,
                 std::vector <OptStruct*> & opt,
                 std::vector <ParStruct*> & par);



#endif
