// Copyright (C) 2015, Antoni Buades, Universitat Illes Balears
// // <toni.buades@uib.es>
// //
// //
// //
// // This program is free software: you can use, modify and/or
// // redistribute it under the terms of the GNU General Public
// // License as published by the Free Software Foundation, either
// // version 3 of the License, or (at your option) any later
// // version. You should have received a copy of this license along
// // this program. If not, see <http://www.gnu.org/licenses/>.
//
//
//

#include "libImage.h"


namespace libUSTG
{



//////////////////////////////////////////////////////////////////
//! Begin Cflimage
//////////////////////////////////////////////////////////////////

//! Constructors
cflimage::cflimage() : d_c(0), d_w(0), d_h(0), d_wh(0), d_whc(0), d_v(0), \
                       visuMin(0.0f), visuMax(255.f)
{
}



cflimage::cflimage(int w, int h, int c) : d_c(c), d_w(w), d_h(h),d_wh(w*h),\
                                          d_whc(c*w*h), d_v(new float[c*w*h]),\
                                          visuMin(0.0f), visuMax(255.f)
{
    for (int ii=0; ii < d_whc; ii++) d_v[ii] = 0.0f;

}



cflimage::cflimage(int w, int h, float *igray) : d_c(1), d_w(w), d_h(h),\
                                                 d_wh(w*h), d_whc(w*h), \
                                                 d_v(new float[w*h]),
                                                 visuMin(0.0f), visuMax(255.f)
{
    memcpy(d_v, igray, w * h * sizeof(float));
}


cflimage::cflimage(int w, int h, unsigned char *igray) : d_c(1),d_w(w),d_h(h),\
                                                         d_wh(w*h), d_whc(w*h),\
                                                         d_v(new float[w*h]), \
                                                         visuMin(0.0f), \
                                                         visuMax(255.f)
{
    for (int ii=0; ii < d_wh; ii++) d_v[ii] = (float) igray[ii];
}


cflimage::cflimage(int w, int h, float *ired, float *igreen, float *iblue) : \
                   d_c(3), d_w(w), d_h(h),d_wh(w*h), d_whc(3*w*h), \
                   d_v(new float[3*w*h]),  visuMin(0.0f), visuMax(255.f)
{
    memcpy(d_v, ired, w * h * sizeof(float));
    memcpy(d_v + w*h, igreen, w * h * sizeof(float));
    memcpy(d_v + 2*w*h, iblue, w * h * sizeof(float));
}


cflimage::cflimage(int w, int h, unsigned char *ired, unsigned char *igreen,
                   unsigned char *iblue) : d_c(3), d_w(w), d_h(h),d_wh(w*h), \
                                           d_whc(3*w*h), d_v(new float[3*w*h]),\
                                           visuMin(0.0f), visuMax(255.f)
{
    for (int ii=0; ii < d_wh; ii++) {
        d_v[ii] = (float) ired[ii];
        d_v[d_wh + ii] = (float) igreen[ii];
        d_v[2*d_wh + ii] = (float) iblue[ii];
    }
}




cflimage::cflimage(const flimage &red, const  flimage &green,
                   const  flimage &blue) : d_c(0), d_w(0), d_h(0), d_wh(0), \
                                           d_whc(0), d_v(0),  visuMin(0.0f), \
                                           visuMax(255.f)
{

    assert(red.d_w == green.d_w && green.d_w == blue.d_w);
    assert(red.d_h == green.d_h && green.d_h == blue.d_h);

    d_w = red.d_w;
    d_h = red.d_h;
    d_c = 3;

    d_wh = d_w * d_h;
    d_whc = d_wh * d_c;

    d_v = new float[3 * d_wh];
    memcpy(d_v, red.d_v, d_wh * sizeof(float));
    memcpy(d_v + d_wh, green.d_v, d_wh * sizeof(float));
    memcpy(d_v + 2 * d_wh, blue.d_v, d_wh * sizeof(float));


}


cflimage::cflimage(const cflimage& im) : d_c(im.d_c), d_w(im.d_w), d_h(im.d_h),\
                                         d_wh(im.d_wh), d_whc(im.d_whc),d_v(0),\
                                         visuMin(0.0f), visuMax(255.f)
{

    if (d_whc > 0) {
        d_v = new float[d_whc];
        memcpy((void *) d_v, (const void *) im.d_v, d_whc * sizeof(float));

        //for (int ii=0; ii < d_whc; ii++) d_v[ii] = im.d_v[ii];
    }

}



void cflimage::create(int w, int h, int c)
{
    erase();
    d_c = c;
    d_w = w;
    d_h=h;
    d_wh = w*h;
    d_whc = c*w*h;
    d_v = new float[d_whc];
    visuMin=0.0f;
    visuMax=255.0f;

    for (int ii=0; ii < d_whc; ii++) d_v[ii] = 0.0f;
}



void cflimage::erase()
{
    d_w = d_h = d_wh = d_whc = 0;
    if (d_v) delete[] d_v;
    d_v=0;
}




cflimage::~cflimage()
{
    erase();
}







//////////////////////////////////////////////////////////////////
//! Begin Operators
//////////////////////////////////////////////////////////////////




cflimage&  cflimage::operator= (const cflimage& im)
{
    if (&im == this) {
        return *this;
    }


    if (d_c != im.d_c || d_w != im.d_w || d_h != im.d_h) {
        erase();
        d_c = im.d_c;
        d_w = im.d_w;
        d_h=im.d_h;
        d_wh = d_w * d_h;
        d_whc=d_c * d_w * d_h;
        d_v = new float[d_whc];
    }


    memcpy((void *) d_v, (const void *) im.d_v, d_whc * sizeof(float));

    visuMin = im.visuMin;
    visuMax = im.visuMax;

    return *this;

}








//////////////////////////////////////////////////////////////////
//! Begin Load/Save
//////////////////////////////////////////////////////////////////



void cflimage::load(const char *filename)
{



    // erase current image
    erase();



    // look if it exists
    std::ifstream fileImage(filename);
    if (!fileImage.good()) {
        std::cout << "... failed to read image " << filename << std::endl;
        exit(-1);

    } else fileImage.close();


    // try pmf format
    d_v = _load_pm(filename, d_c, d_w, d_h);
    if (d_v) {

        d_wh = d_w * d_h;
        d_whc = d_c * d_w * d_h;
        return;
    }



    // the sure bet
    d_v = iio_read_image_float_split(filename, &d_w, &d_h, &d_c);

    if (d_v) {
        //if (d_c == 2) d_c = 1;
        //if (d_c > 3)  d_c = 3;

        d_wh = d_w * d_h;
        d_whc = d_c * d_w * d_h;

        return;
    }




    std::cout << "... failed to read image " << filename << std::endl;
    exit(-1);


}



void cflimage::save(const char *filename)
{

    if (!d_v) {
        std::cout << "... failed to save image " << filename << std::endl;
        exit(-1);
    }

    std::string strname(filename);
    size_t pos = strname.find_last_of ('.');
    std::string extension = strname.substr(pos+1);

    if ( extension == "png" || extension == "PNG" || extension == "tif" || \
        extension == "TIF" || extension == "tiff" || extension == "TIFF" ) {

        iio_save_image_float_split((char*)filename, d_v, d_w, d_h, d_c);

    } else {

        if (_create_pm(filename,d_c,d_w,d_h,d_v) == 0) {
          std::cout << "... failed to save pmf image " << filename << std::endl;
          exit(-1);

        } else return;

    }

}






//////////////////////////////////////////////////////////////////
//! Begin Get Basic Data
//////////////////////////////////////////////////////////////////




cflimage::operator  flimage()
{
    return getGray();
}





int  cflimage::isSameSize(const  cflimage &inIm)
{

    if (d_c != inIm.d_c || d_w != inIm.d_w || d_h != inIm.d_h) return 0;
    else return 1;

}






flimage cflimage::getChannel(int i)
{

    assert(i < d_c);

    flimage image(d_w,d_h);

    for (int jj=0; jj < d_wh; jj++) image.d_v[jj] = d_v[ i * d_wh + jj];

    return image;
}



void cflimage::getChannel(int i, flimage *out)
{

    assert(i < d_c);
    assert(d_v != NULL);

    if (out->d_w != d_w || out->d_h != d_h) {
        out->erase();
        out->create(d_w, d_h);
    }


    for (int jj=0; jj < d_wh; jj++) out->d_v[jj] = d_v[ i * d_wh + jj];


}




//////////////////////////////////////////////////////////////////
//! Begin Math
//////////////////////////////////////////////////////////////////



cflimage& cflimage::operator= (float a)
{
    if (d_v)  for (int j=0; j<d_whc ; j++) d_v[j] = a;

    return *this;
}


void cflimage::operator-= (float a)
{
    if (d_v)  for (int j=0; j<d_whc ; j++) d_v[j] -= a;

}


void cflimage::operator+= (float a)
{
    if (d_v)  for (int j=0; j<d_whc ; j++) d_v[j] += a;

}


void cflimage::operator*= (float a)
{
    if (d_v)  for (int j=0; j<d_whc ; j++) d_v[j] *= a;

}




float  cflimage::max ()
{

    assert(d_v != NULL);

    float fmax = d_v[0];
    for (int j=0; j < d_whc ; j++)  if (d_v[j] > fmax)  fmax = d_v[j];
    return fmax;


}



float  cflimage::min ()
{
    assert(d_v != NULL);

    float fmin = d_v[0];
    for (int j=0; j < d_whc ; j++)  if (d_v[j] < fmin)  fmin = d_v[j];
    return fmin;
}



float  cflimage::min_channel (int i)
{
    assert(d_v != NULL);
    assert(i>= 0 && i < d_c);

    float *ptr = &d_v[i * d_wh];
    float fmin = *ptr;

    for (int j=0; j < d_wh ; j++, ptr++)  if (*ptr < fmin)  fmin = *ptr;
    return fmin;
}



float  cflimage::max_channel (int i)
{
    assert(d_v != NULL);
    assert(i>= 0 && i < d_c);

    float *ptr = &d_v[i * d_wh];
    float fmax = *ptr;

    for (int j=0; j < d_wh ; j++, ptr++)  if (*ptr > fmax)  fmax = *ptr;
    return fmax;
}



void  cflimage::min (float m)
{
    assert(d_v != NULL);
    for (int j=0; j < d_whc ; j++)  if (d_v[j] > m)  d_v[j] = m;
}



void  cflimage::max (float M)
{
    assert(d_v != NULL);
    for (int j=0; j < d_whc ; j++)  if (d_v[j] < M)  d_v[j] = M;
}




void cflimage::thre(float m, float M)
{

    assert(d_v != NULL);
    for (int ii=0; ii < d_whc; ii++) {
        if (d_v[ii] >= M) 	d_v[ii]= M;
        else if (d_v[ii] <= m)  d_v[ii]= m;
    }

}



void cflimage::normalizeL1()
{

    float fSum = 0.0f;
    for (int j=0; j < d_whc ; j++)    fSum += d_v[j];

    assert(fSum != 0.0f);
    float dfSum = 1.0 / fSum;
    for (int j=0; j < d_whc ; j++)    d_v[j] *= dfSum;

}



void cflimage::rint()
{

    assert(d_v != NULL);
    for (int ii=0; ii < d_whc; ii++)		d_v[ii]= rintf(d_v[ii]);

}


void cflimage::abs()
{

    assert(d_v != NULL);
    for (int ii=0; ii < d_whc; ii++)		d_v[ii]= fabsf(d_v[ii]);

}





///////////////////////////////////////////////
//! Begin Color Conversion
///////////////////////////////////////////////


flimage cflimage::getGray()
{

    flimage image(d_w, d_h);
    image=0.0f;

    getGray(&image);

    return image;
}





void cflimage::getGray(flimage * out)
{



    assert(d_v != NULL);
    if (out->d_w != d_w || out->d_h != d_h) {
        out->erase();
        out->create(d_w, d_h);
    }


    for (int i=0; i < d_whc; i++) {
        out->d_v[i % d_wh] += d_v[i];
    }

    for (int i=0; i < d_wh; i++)
        out->d_v[i] /= (float) d_c;

}



flimage cflimage::getGray(float wr, float wg, float wb)
{

    //assert(d_c == 1  || d_c == 3);

    flimage image(d_w, d_h);
    image=0.0f;

    getGray(wr,wg,wb,&image);

    return image;
}





void cflimage::getGray(float wr, float wg, float wb, flimage *out)
{
    //assert(d_c == 1  || d_c == 3);
    assert(d_v != NULL);

    if (out->d_w != d_w || out->d_h != d_h) {
        out->erase();
        out->create(d_w, d_h);
    }


    if (d_c == 1)  *out = (flimage) (*this);
    else {
        for (int i=0; i < d_wh; i++) out->d_v[i] = wr * d_v[i] + \
                                                   wg * d_v[d_wh + i] + \
                                                   wb * d_v[2*d_wh+i];
    }


}




cflimage  cflimage::gradient(char cType)
{
    assert(d_v != NULL);

    cflimage grad(d_w, d_h, d_c);

    for (int ii=0; ii < d_c; ii++) {
        fiComputeImageGradient(&d_v[ii * d_wh], &grad.d_v[ii * d_wh],
                               NULL, d_w, d_h,cType);
    }

    return grad;
}



cflimage  cflimage::xgradient(char cType)
{
    assert(d_v != NULL);

    cflimage grad(d_w, d_h, d_c);

    for (int ii=0; ii < d_c; ii++) {
        fiComputeImageGradient(&d_v[ii * d_wh], &grad.d_v[ii * d_wh], NULL,
                               NULL, NULL, d_w, d_h,cType);
    }

    return grad;
}



cflimage  cflimage::ygradient(char cType)
{
    assert(d_v != NULL);

    cflimage grad(d_w, d_h, d_c);

    for (int ii=0; ii < d_c; ii++) {
        fiComputeImageGradient(&d_v[ii * d_wh], NULL, &grad.d_v[ii * d_wh],
                               NULL, NULL, d_w, d_h,cType);
    }

    return grad;
}




cflimage  cflimage::gradient(cflimage &orientation, char cType)
{
    assert(d_v != NULL);

    cflimage grad(d_w, d_h, d_c);
    if (!orientation.isSameSize(grad)) {
        orientation.erase();
        orientation.create(d_w, d_h, d_c);
    }

    for (int ii=0; ii < d_c; ii++) {
        fiComputeImageGradient(&d_v[ii * d_wh], &grad.d_v[ii * d_wh],
                               &orientation.d_v[ii * d_wh], d_w, d_h,cType);
    }

    return grad;
}



cflimage  cflimage::gradient(cflimage &xgrad, cflimage &ygrad,
                             cflimage &orientation, char cType)
{
    assert(d_v != NULL);

    cflimage grad(d_w, d_h, d_c);
    if (!orientation.isSameSize(grad)) {
        orientation.erase();
        orientation.create(d_w, d_h, d_c);
    }
    if (!xgrad.isSameSize(grad)) {
        xgrad.erase();
        xgrad.create(d_w, d_h, d_c);
    }
    if (!ygrad.isSameSize(grad)) {
        ygrad.erase();
        ygrad.create(d_w, d_h, d_c);
    }

    for (int ii=0; ii < d_c; ii++) {
        fiComputeImageGradient(&d_v[ii * d_wh], &xgrad.d_v[ii * d_wh],
                               &ygrad.d_v[ii * d_wh], &grad.d_v[ii * d_wh],
                               &orientation.d_v[ii * d_wh], d_w, d_h,cType);
    }

    return grad;
}









///////////////////////////////////////////////
//! Begin Sampling and Convolution
///////////////////////////////////////////////


void cflimage::convolve_skernel(struct sorted_kernel *skernel, int boundary,
                                cflimage *out)
{
    assert(d_v != NULL);
    if (!isSameSize(*out)) {
        out->erase();
        out->create(d_w, d_h, d_c);
    }

    for (int i=0; i < d_c; i++) {
        fiConvol_skernel( &d_v[i*d_wh], &(out->d_v[i*d_wh]), d_w, d_h,
                         skernel, boundary);

    }

}




/// Class flimage

flimage::flimage() : cflimage()
{
}




flimage::flimage(int w, int h) : cflimage(w, h, 1)
{
}


flimage::flimage(int w, int h, float *ptr) : cflimage(w, h, 1)
{
    memcpy(this->d_v, ptr, w * h * sizeof(float));
}

flimage::flimage(int w, int h, unsigned char *ptr) : cflimage(w, h, ptr)
{
}




flimage::flimage(const flimage& im)
    : cflimage(im)
{}




flimage& flimage::operator=(const flimage & im)
{
    cflimage::operator=(im);
    return *this;
}




void flimage::create(int w, int h)
{
    cflimage::create(w,h,1);
}





}


